#include "xwin.h"

static XColor xwinquery[256];
static XColor xwinanswer[256];
static XColor xwininternal[256];

void xwinflushmap(txwin *xw, int start, int end)
{
  XStoreColors(xw->display, xw->colormap, xwininternal+start, (end - start) + 1);
}

void xwincreatespec()
{
  int count=0,red=0,green=0,blue=0;

  for (blue = 35; blue < 250; blue += 6)
  {
    xwininternal[count].red   = 0;
    xwininternal[count].green = 0;
    xwininternal[count].blue  = blue << 8;
    count++;
  }
  for (green = 35; green < 250; green += 6)
  {
    xwininternal[count].red   = 0;
    xwininternal[count].green = green << 8;
    xwininternal[count].blue  = blue  << 8;
    blue -= 6;
    count++;
  }
  for (blue = 35; blue < 250; blue += 6)
  {
    xwininternal[count].red   = 0;
    xwininternal[count].green = green << 8;
    xwininternal[count].blue  = blue  << 8;
    count++;
  }
  for (red = 35; red < 250; red += 6)
  {
    xwininternal[count].red   = red   << 8;
    xwininternal[count].green = green << 8;
    xwininternal[count].blue  = blue  << 8;
    green -= 6;
    blue  -= 6;
    count++;
  }
  for (blue = 35; blue < 250; blue += 6)
  {
    xwininternal[count].red   = red   << 8;
    xwininternal[count].green = 0;
    xwininternal[count].blue  = blue  << 8;
    count++;
  }
  for (green = 35; green < 250; green += 6)
  {
    xwininternal[count].red   = red   << 8;
    xwininternal[count].green = green << 8;
    xwininternal[count].blue  = blue  << 8;
    blue -= 6;
    count++;
  }
  for (blue = 35; blue < 250; blue += 6)
  {
    xwininternal[count].red   = red   << 8;
    xwininternal[count].green = green << 8;
    xwininternal[count].blue  = blue  << 8;
    count++;
  }
}

txwin xwincreate(char *dispname)
{
  txwin new;
  
  int index;
  for (index = 0; index < 256; index++)
  {
    xwinanswer[index].pixel   = index;
    xwinquery[index].pixel    = index;
    xwininternal[index].pixel = index;

    xwinanswer[index].flags   = DoRed | DoGreen | DoBlue;
    xwinquery[index].flags    = DoRed | DoGreen | DoBlue;
    xwininternal[index].flags = DoRed | DoGreen | DoBlue;
  }

  new.display = XOpenDisplay(dispname);
  
  if (new.display)
  {
    new.screen      = DefaultScreen(new.display);
    new.root        = DefaultRootWindow(new.display);
    new.depth       = DefaultDepth(new.display,new.screen);
    new.visual      = DefaultVisual(new.display,new.screen);
    new.colormap    = XCreateColormap(new.display,new.root,new.visual,AllocAll);
    new.currentfont = XLoadFont(new.display,"7x14");
    xwincreatespec();
    xwinflushmap(&new, 0, 255);
  }
  return(new);
}

void xwinremove(txwin *xw)
{
  XCloseDisplay(xw->display);
}

Window xwinopen(txwin *xw, char *windname,char *iconname,int width,
 int height,int maxwidth,int maxheight,int minwidth,int minheight,
 unsigned int class,unsigned long background,int xpos,int ypos)
{
  XSizeHints            hints;
  XWMHints              xwmh;
  XEvent                event;
  unsigned long         event_mask;
  Window                new;
  XSetWindowAttributes  winattr;	
  unsigned long         valuemask;	

  valuemask = CWBackPixel | CWBorderPixel | CWColormap;

  winattr.background_pixel = background;
  winattr.border_pixel     = xw->foreground;
  winattr.colormap         = xw->colormap;

  new = XCreateWindow(xw->display,xw->root,xpos,ypos,width,height,1,xw->depth,class,xw->visual,valuemask,&winattr);

  hints.flags       = PSize | PMinSize | PMaxSize | PPosition | PWinGravity;
  hints.x           = xpos;
  hints.y           = ypos;
  hints.width       = width;
  hints.height      = height;
  hints.min_width   = minwidth;
  hints.max_width   = maxwidth;
  hints.min_height  = minheight;
  hints.max_height  = maxheight;
  hints.win_gravity = CenterGravity;

  XSetStandardProperties(xw->display, new, windname, iconname, None, 0, 0, &hints);

  xwmh.input         = True;
  xwmh.initial_state = NormalState;
  xwmh.flags         = InputHint | StateHint;

  XSetWMHints(xw->display, new, &xwmh);

  event_mask = ButtonPressMask | NoEventMask | ButtonReleaseMask | KeyPressMask | EnterWindowMask | KeyReleaseMask | LeaveWindowMask
| ExposureMask | PointerMotionMask | VisibilityChangeMask | StructureNotifyMask | Button1MotionMask | ResizeRedirectMask | Button2MotionMask
| SubstructureNotifyMask | Button3MotionMask | SubstructureRedirectMask | Button4MotionMask | FocusChangeMask | Button5MotionMask | PropertyChangeMask
| ButtonMotionMask | ColormapChangeMask | KeymapStateMask | OwnerGrabButtonMask;

  XSelectInput(xw->display, new, event_mask);

  XMapWindow(xw->display, new);

  while (True)
  {
    XNextEvent(xw->display, &event);
    if (event.xany.window != new)
      continue;

    if (event.type == Expose) 
      break;
  } 
  XFlush(xw->display);

  return (new);
}

void xwinclose(txwin *xw, Window old)
{
  XDestroyWindow(xw->display, old);
}

GC xwinopengc(txwin *xw, Window window)
{
XGCValues      gcv;
unsigned long  mask;
GC             new;					

  mask = GCForeground | GCBackground | GCFunction | GCFont;

  gcv.foreground = xw->foreground;
  gcv.background = xw->background;
  gcv.font       = xw->currentfont;
  gcv.function   = GXcopy;

  new = XCreateGC(xw->display, window, mask, &gcv);

  return(new);
}

void xwinfreegc(txwin *xw, GC old)
{
  XFreeGC(xw->display, old);
}

Pixmap xwinopenpix(txwin *xw, int width, int height)
{
  Pixmap new;

  new = XCreatePixmap(xw->display,xw->root,width,height,DefaultDepth(xw->display,xw->screen));

  return(new);
}	

void xwinclosepix(txwin *xw, Pixmap old)
{
  XFreePixmap(xw->display, old);
}

void xwincycledn(int start, int end)
{
  int           index;
  int           colors = end-start+1;
  int           factor = 0;

  for (index = start; index <= end; index++)
  {
     factor = start+(index+1-start)%colors;

     xwininternal[index].red   = xwininternal[factor].red;
     xwininternal[index].green = xwininternal[factor].green;
     xwininternal[index].blue  = xwininternal[factor].blue;
  }
}

void xwincycleup(int start, int end)
{
  int    index;
  int    factor = 0;
  int    red;
  int    green;
  int    blue;

  red   = xwininternal[end].red;
  green = xwininternal[end].green;
  blue  = xwininternal[end].blue;  

  xwininternal[start].red   = xwininternal[end].red;
  xwininternal[start].green = xwininternal[end].green;
  xwininternal[start].blue  = xwininternal[end].blue;

  for (index = end; index > start; index--)
  {
     factor = index-1;
     xwininternal[index].red   = xwininternal[factor].red;
     xwininternal[index].green = xwininternal[factor].green;
     xwininternal[index].blue  = xwininternal[factor].blue;
  }

  xwininternal[start].red   = red;
  xwininternal[start].green = green;
  xwininternal[start].blue  = blue;
}

void xwinfadedn(int start, int end)
{
  int index;

  for (index = start; index < end; index++)
  {
     xwininternal[index].red   = (xwininternal[index].red-256   < 0) ? 0 : xwininternal[index].red-256;
     xwininternal[index].green = (xwininternal[index].green-256 < 0) ? 0 : xwininternal[index].green-256;
     xwininternal[index].blue  = (xwininternal[index].blue-256  < 0) ? 0 : xwininternal[index].blue-256;
  }
}

void xwinloadmap(txwin *xw, Colormap colormap, int start, int end)
{
   XQueryColors(xw->display, colormap, xwininternal+start, (end - start) + 1);
}

void xwinrgb2col(int colornumber, int red, int green, int blue)
{
  xwininternal[colornumber].red   = red   << 8;
  xwininternal[colornumber].green = green << 8;
  xwininternal[colornumber].blue  = blue  << 8;
}
